/*
 * dctpRawApi.c -- This file implements the primary interface between the dccp protocl stack
 * and the application level.  The routines here normally won't be called directly by the
 * applications.  There will normally be some environment-specific API that then calls these
 * routines.
 *
 * Functions whose names begin with 'dctpa' are meant to be called by the environment-
 * specific API.  They follow the usual socket conventions, with optimizations specific to
 * the dccp-tp stack.  Applications first use dctaSocket to create the socket.  Applications
 * must supply a DCCP Service Code for the socket at this time.
 *
 * Client applications then use dctpaConnect to connect to a server, dctpaSend and dctpaRecv
 * to send and receive data and dctpaClose to close the socket when things are all done.
 * Unlike the normal 'connect' call, dctpaConnect will take user data to add to the
 * DCCP-Request packet.
 *
 * After creating the socket, server applications use dctpaBind to supply a local address and
 * port and then listen on that port/service code combination with dctpaListen.  Unlike
 * other transport protocols, multiple sockets may listen to the same port as long as the
 * service code associated with each is unique.  Connection requests arriving for a listening
 * server can be accepted with dctpaAccept.  Data received in the DCCP-Request (if any) is
 * delivered in the first call to dctpaRecv.  On the client side, data received in the
 * DCCP-Response is also delivered by the first call to dctpaRecv.
 *
 * Copyright (C) 2008 Tom Phelan
 *
 * This file is part of dccp-tp.
 *
 * Dccp-tp is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * Dccp-tp is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with dccp-tp.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Documentation and source code for dccp-tp is available at
 * http://www.phelan-4.com/dccp-tp/.
 */

#include "dctpCore.h"
#include "dctpRawApi.h"
#include "dctpSupport.h"
#include "dctpInternal.h"

/*
 * Prototypes for private functions
 */
static int getPortListPointers(uint_t encap, uint8_t **plist, uint_t **fport,
			       dctpSupportMutex_t **plock);
static int getFreePort(uint_t encap);

/*
 * badsock -- private routine to check socket validity.
 */
static int badsock(dctpSocket *sock, int *cerrno) {
    if ((sock == NULL) || (sock->magic != DCTPSOCK_MAGICNO)) {
	if (cerrno) *cerrno = DCTP_ENOTSOCK;
	return(-1);
    }
    return(0);
}

/*
 * badaddr -- private routine to check address validity
 */

static int badaddr(dctpSocket *sock, dctpSockaddr *addr, int *cerrno) {
    if (addr) {
	switch (sock->encap) {
	case DCTPENCAP_RAW:
	case DCTPENCAP_NAT:
	    if (addr->sa_family != DCTP_AF_INET) {
		if (cerrno) *cerrno = DCTP_EINVAL;
		return(-1);
	    }
	    return(0);
	case DCTPENCAP_RAW6:
	case DCTPENCAP_NAT6:
	    if (addr->sa_family != DCTP_AF_INET6) {
		if (cerrno) *cerrno = DCTP_EINVAL;
		return(-1);
	    }
	    return(0);
	default:
	    if (cerrno) *cerrno = DCTP_EBADFD;
	    return(-1);
	}
    } else {
	return(-1);
    }
}

/*
 * addrcpy -- private routine to support address copying
 */

static void addrcpy(dctpSockaddr *dest, dctpSockaddr *src) {
    dctpoMemcpy(dest, src, (src->sa_family == DCTP_AF_INET) ?
		sizeof(dctpSockaddrV4) : sizeof(dctpSockaddrV6));
}

/*
 * dctpaSocket -- Create a DCCP socket.
 *
 * Inputs:  encap --  one of DCTPENCAP_RAW for IPv4 direct
 *                           DCTPENCAP_NAT for IPv4 over UDP
 *                           DCTPENCAP_RAW6 for IPv6 direct
 *                           DCTPENCAP_NAT6 for Ipv6 over UDPv6
 *          scode --  Service Code for this socket.  MUST NOT be 0xffffffff,
 *                    SHOULD NOT be 0.  In network byte order.
 *          cerrno -- Pointer to errno-style variable for error details.
 *                    May be NULL if error details not needed.
 *
 * Returns: Pointer to dctpaSocket for success
 *          NULL for error, with *cerrno set (if not NULL) to extended error.
 *          Extended errors mimick Unix 'socket' call.
 */
dctpSocket *dctpaSocket(int encap, uint32_t scode, int *cerrno) {
    dctpSocket *sock;

    if (scode == 0xffffffff) {
	/* Must be valid service code */
	if (cerrno) *cerrno = DCTP_EINVAL;
	return(NULL);
    }
    switch (encap) {
    case DCTPENCAP_RAW:
    case DCTPENCAP_RAW6:
    case DCTPENCAP_NAT:
    case DCTPENCAP_NAT6:
	break;
    default:
	if (cerrno) *cerrno = DCTP_EINVAL;
	return(NULL);
    }
    if ((sock = dctpoMalloc(sizeof(dctpSocket))) == NULL) {
	if (cerrno) *cerrno = DCTP_ENOMEM;
	return(NULL);
    }
    dctpoMemset(sock, 0, sizeof(dctpSocket));
    /* Set things that default to non-zero */
    sock->magic = DCTPSOCK_MAGICNO;
    sock->state = DCTPSOCK_CREATED;
    sock->scode = scode;
    sock->encap = encap;
    sock->mxpktqlen = DCTPSOCK_DEF_MAXPKTQLEN;
    sock->mrpktqlen = DCTPSOCK_DEF_MAXPKTQLEN;
    /* Current local features */
    sock->localFeatsCurr.ccid[0] = DCTPFEAT_CCIDDEF;
    sock->localFeatsCurr.ackRatio = DCTPFEAT_ACKRATIONDEF;
    sock->localFeatsCurr.seqWindow = DCTPFEAT_SEQWINDOWDEF;
    /* Desired local features */
    sock->localFeatsDesired.ccid[0] = DCTPFEAT_CCIDDEF;
    sock->localFeatsDesired.ackRatio = DCTPFEAT_ACKRATIONDEF;
    sock->localFeatsDesired.seqWindow = DCTPFEAT_SEQWINDOWDEF;
    /* Current remote features */
    sock->remoteFeatsCurr.ccid[0] = DCTPFEAT_CCIDDEF;
    sock->remoteFeatsCurr.ackRatio = DCTPFEAT_ACKRATIONDEF;
    sock->remoteFeatsCurr.seqWindow = DCTPFEAT_SEQWINDOWDEF;
    /* Desired remote features */
    sock->remoteFeatsDesired.ccid[0] = DCTPFEAT_CCIDDEF;
    sock->remoteFeatsDesired.ackRatio = DCTPFEAT_ACKRATIONDEF;
    sock->remoteFeatsDesired.seqWindow = DCTPFEAT_SEQWINDOWDEF;
    /* CCID 2 requires that we start with negotiating Send Ack Vector */
    sock->remoteFeatsDesired.sendAckVector = 1;
    sock->remoteFeatsState[DCTPFEAT_SENDACKVECTOR] = DCTPFEAT_MANUNSTABLE;
    sock->localFeatsDesired.sendAckVector = 1;
    sock->localFeatsState[DCTPFEAT_SENDACKVECTOR] = DCTPFEAT_MANUNSTABLE;
    sock->sendFeatureNegotiation = 1;
    dctpoSocketLockInit(sock);
    return(sock);
}

/*
 * dctpaBind -- Bind a local address and port to a socket.
 *
 * Inputs:  sock --   pointer to socket in CREATED state.
 *          addr --   pointer to local address/port.  MUST be in proper format
 *                    for socket's encapsulation (AF_INET or AF_INET6).
 *          cerrno -- Pointer to errno-style variable for error details.
 *                    May be NULL if error details not needed.
 *
 * Returns: Zero (0) on success.
 *          -1 for error, with *cerrno set (if not NULL) to extended error.
 *          Extended errors mimick Unix 'bind' call.
 */
int dctpaBind(dctpSocket *sock, dctpSockaddr *addr, int *cerrno) {
    if (badsock(sock, cerrno)) return(-1);
    if (badaddr(sock, addr, cerrno)) return(-1);
    dctpoSocketLock(sock);
    if (sock->state != DCTPSOCK_CREATED) {
	dctpoSocketUnlock(sock);
	if (cerrno) *cerrno = DCTP_EINVAL;
	return(-1);
    }
    addrcpy(&(sock->localAddr), addr);
    sock->state = DCTPSOCK_BOUND;
    dctpoSocketUnlock(sock);
    return(0);
}

/*
 * dctpaListen -- Listen for connections on a socket
 *
 * Inputs:  sock --    pointer to socket in BOUND state.
 *          backlog -- Maximum number of waiting connections to enqueue.  Silently
 *                     reduced to DCTP_MAX_BACKLOG if greater.
 *          cerrno --  Pointer to errno-style variable for error details.
 *                     May be NULL if error details not needed.
 *
 * Returns: Zero (0) on success.
 *          -1 for error, with *cerrno set (if not NULL) to extended error.
 *          Extended errors mimick Unix 'listen' call.
 */
int dctpaListen(dctpSocket *sock, unsigned int backlog, int *cerrno) {
    if (badsock(sock, cerrno)) return(-1);
    dctpoSocketLock(sock);
    if (sock->state != DCTPSOCK_BOUND) {
	dctpoSocketUnlock(sock);
	if (cerrno) *cerrno = DCTP_EINVAL;
	return(-1);
    }
    if (dctpiGetListenSock(sock) != NULL) {
	dctpoSocketUnlock(sock);
	if (cerrno) *cerrno = DCTP_EADDRINUSE;
	return(-1);
    }
    sock->backlog = (backlog > DCTPSOCK_MAXBACKLOG) ? DCTPSOCK_MAXBACKLOG : backlog;
    sock->state = DCTPSOCK_LISTEN;
    sock->server = 1;
    dctpiAddListenSock(sock);
    dctpoSocketUnlock(sock);
    return(0);
}

/*
 * dctpaAccept -- accept a connection on a socket
 *
 * Inputs:  sock --   pointer to socket in LISTEN state.
 *          addr --   remote address copied here on succesful connection. 
 *          alen --   length of addr.
 *          cerrno -- Pointer to errno-style variable for error details.
 *                    May be NULL if error details not needed.
 *
 * Returns: Zero (0) on success.
 *          -1 for error, with *cerrno set (if not NULL) to extended error.
 *          Extended errors mimick Unix 'accept' call.
 */
dctpSocket *dctpaAccept(dctpSocket *sock, dctpSockaddr *addr, int alen, int *cerrno) {
    dctpSocket *nsock;

    if (badsock(sock, cerrno)) return(NULL);
    dctpoSocketLock(sock);
    if (sock->state != DCTPSOCK_LISTEN) {
	dctpoSocketUnlock(sock);
	if (cerrno) *cerrno = DCTP_EINVAL;
	return(NULL);
    }
    while ((nsock = dctpiGetWaitingSock(sock)) == NULL) {
	if (sock->flags & DCTPSOCK_O_NONBLOCK) {
	    if (cerrno) *cerrno = DCTP_EWOULDBLOCK;
	    return(NULL);
	}
	dctpoSocketWait(sock);
	if (sock->state != DCTPSOCK_LISTEN) {
	    dctpoSocketUnlock(sock);
	    if (cerrno) *cerrno = DCTP_ECONNREFUSED;
	    return(NULL);
	}
    }
    dctpoSocketUnlock(sock);
    dctpoMemcpy(addr, &(nsock->remoteAddr), alen);
    return(nsock);
}

/*
 * Keep track of ports used by client apps
 */
static dctpSupportMutex_t natV4FreePortLock = DCTP_OBJECTLOCKINIT;
static uint8_t natV4FreePorts[8192];       /* Each bit represents a port */
static uint_t nxtNatV4FreePort = DCTP_DYNAMICPORTSTART;
static dctpSupportMutex_t rawV4FreePortLock = DCTP_OBJECTLOCKINIT;
static uint8_t rawV4FreePorts[8192];       /* Each bit represents a port */
static uint_t nxtRawV4FreePort = DCTP_DYNAMICPORTSTART;
static dctpSupportMutex_t natV6FreePortLock = DCTP_OBJECTLOCKINIT;
static uint8_t natV6FreePorts[8192];       /* Each bit represents a port */
static uint_t nxtNatV6FreePort = DCTP_DYNAMICPORTSTART;
static dctpSupportMutex_t rawV6FreePortLock = DCTP_OBJECTLOCKINIT;
static uint8_t rawV6FreePorts[8192];       /* Each bit represents a port */
static uint_t nxtRawV6FreePort = DCTP_DYNAMICPORTSTART;

/*
 * Get proper set of pointers for encap
 */
static int getPortListPointers(uint_t encap, uint8_t **plist, uint_t **fport,
			       dctpSupportMutex_t **plock) {
    switch (encap) {
    case DCTPENCAP_RAW:
	*plist = rawV4FreePorts;
	*fport = &nxtRawV4FreePort;
	*plock = &rawV4FreePortLock;
	return(0);
    case DCTPENCAP_NAT:
	*plist = natV4FreePorts;
	*fport = &nxtNatV4FreePort;
	*plock = &natV4FreePortLock;
	return(0);
    case DCTPENCAP_RAW6:
	*plist = rawV6FreePorts;
	*fport = &nxtRawV6FreePort;
	*plock = &rawV6FreePortLock;
	return(0);
    case DCTPENCAP_NAT6:
	*plist = natV6FreePorts;
	*fport = &nxtNatV6FreePort;
	*plock = &natV6FreePortLock;
	return(0);
    default:
	return(-1);
    }
}

/*
 * Find a free port
 */
static int getFreePort(uint_t encap) {
    uint8_t *plist;
    uint_t *fport, fpstart;
    dctpSupportMutex_t *plock;

    if (getPortListPointers(encap, &plist, &fport, &plock) < 0) {
	return(-1);
    }
    dctpoObjectLock(plock);
    fpstart = *fport;
    while ((plist[*fport >> 3] & (1 << (*fport & 7))) != 0) {
	if (++(*fport) > 65535) *fport = DCTP_DYNAMICPORTSTART;
	if (*fport == fpstart) {
	    dctpoObjectUnlock(plock);
	    return(-1);
	}
    }
    fpstart = *fport;
    plist[fpstart >> 3] |= 1 << (fpstart & 7);
    if (++(*fport) > 65535) *fport = 1;
    dctpoObjectUnlock(plock);
    fpstart = dctpoHtons(fpstart) & 0xffff;
    return(fpstart);
}

/*
 * Return port to free
 */
int dctpiFreePort(uint_t encap, uint16_t port) {
    uint8_t *plist;
    uint_t *fport;
    dctpSupportMutex_t *plock;

    port = dctpoNtohs(port);
    if (getPortListPointers(encap, &plist, &fport, &plock) < 0) {
	return(-1);
    }
    dctpoObjectLock(plock);
    plist[port >> 3] &= ~(1 << (port & 7));
    dctpoObjectUnlock(plock);
    return(0);
}

/*
 * dctpaConnect -- initiate a connection on a socket
 *
 * Inputs:  sock --   pointer to socket in CREATED state.
 *          addr --   remote address to connect to.
 *          data --   packet with application data for DCCP-Request.  May be NULL.
 *          cerrno -- Pointer to errno-style variable for error details.
 *                    May be NULL if error details not needed.
 *
 * Returns: Zero (0) on success.
 *          -1 for error, with *cerrno set (if not NULL) to extended error.
 *          Extended errors mimick Unix 'connect' call.
 */
int dctpaConnect(dctpSocket *sock, dctpSockaddr *addr, dctpPacket *data, int *cerrno) {
    int sport;

    if (badsock(sock, cerrno)) return(-1);
    if (badaddr(sock, addr, cerrno)) return(-1);
    dctpoSocketLock(sock);
    if (sock->state != DCTPSOCK_CREATED) {
	dctpoSocketUnlock(sock);
	if (cerrno) *cerrno = DCTP_EINVAL;
	return(-1);
    }
    addrcpy(&(sock->remoteAddr), addr);
    if ((sport = getFreePort(sock->encap)) < 0) {
	dctpoSocketUnlock(sock);
	if (cerrno) *cerrno = DCTP_EADDRINUSE;
	return(-1);
    }
    sock->localAddr.sa_port = sport;
    dctpiAddConnectedSocket(sock);
    sock->state = DCTPSOCK_REQUEST;
    sock->server = 0;
    sock->iss = dctpoRandomSeqno();
    sock->gss = sock->iss;
    sock->awl = sock->iss;
    sock->awlinit = 1;
    sock->awh = sock->iss;
    sock->fgss = sock->iss;
    dctpiSendRequest(sock, data);
    sock->timeval = DCTP_INITIALRQTIMEOUT;
    dctpoStartTimer(&(sock->timer), sock->timeval, dctpiRequestTimeout, sock);
    /* Wait for response */
    while (sock->state == DCTPSOCK_REQUEST) {
	if (sock->flags & DCTPSOCK_O_NONBLOCK) {
	    if (cerrno) *cerrno = DCTP_EWOULDBLOCK;
	    dctpoSocketUnlock(sock);
	    return(-1);
	}
	dctpoSocketWait(sock);
    }
    if ((sock->state != DCTPSOCK_PARTOPEN) && (sock->state != DCTPSOCK_OPEN))  {
	if (cerrno) *cerrno = DCTP_ECONNREFUSED;
	dctpoSocketUnlock(sock);
	return(-1);
    }
    dctpoSocketUnlock(sock);
    return(0);
}

/*
 * dctpaSend -- Send data on a connection
 *
 * Inputs:  sock --   pointer to socket in OPEN or PARTOPEN state
 *          data --   packet with data to send.  Zero-length data sent as
 *                    zero-length packet.
 *          cerrno -- Pointer to errno-style variable for error details.
 *                    May be NULL if error details not needed.
 *
 * Returns: Zero (0) on success.
 *          -1 for error, with *cerrno set (if not NULL) to extended error.
 *          Extended errors mimick Unix call.
 */
int dctpaSend(dctpSocket *sock, dctpPacket *data, int *cerrno) {
    int dlen = data->appdatalen;

    if (badsock(sock, cerrno)) return(-1);
    dctpoSocketLock(sock);
    if ((sock->state != DCTPSOCK_PARTOPEN) && (sock->state != DCTPSOCK_OPEN)) {
	if (cerrno) *cerrno = DCTP_EINVAL;
	dctpoSocketUnlock(sock);
	return(-1);
    }
    while (sock->xpktqlen >= sock->mxpktqlen) {
	if (sock->flags & DCTPSOCK_O_NONBLOCK) {
	    if (cerrno) *cerrno = DCTP_EWOULDBLOCK;
	    dctpoSocketUnlock(sock);
	    return(-1);
	}
	dctpoSocketWait(sock);
	if ((sock->state != DCTPSOCK_PARTOPEN) && (sock->state != DCTPSOCK_OPEN)) {
	    if (cerrno) *cerrno = DCTP_ECONNREFUSED;
	    dctpoSocketUnlock(sock);
	    return(-1);
	}
    }
    data->nextPkt = NULL;
    if (sock->lastXpkt) {
	sock->lastXpkt->nextPkt = data;
	sock->lastXpkt = data;
    } else {
	sock->firstXpkt = data;
	sock->lastXpkt = data;
    }
    ++sock->xpktqlen;
    dctpiXmtPacket(sock);
    dctpoSocketUnlock(sock);
    return(dlen);
}

/*
 * dctpaRecv -- Receive data on a connection
 *
 * Inputs:  sock --   pointer to socket in OPEN or PARTOPEN state
 *          data --   place to put pointer to received data
 *          cerrno -- Pointer to errno-style variable for error details.
 *                    May be NULL if error details not needed.
 *
 * Returns: Zero (0) on success.
 *          -1 for error, with *cerrno set (if not NULL) to extended error.
 *          Extended errors mimick Unix call.
 */
int dctpaRecv(dctpSocket *sock, dctpPacket **data, int *cerrno) {
    if (badsock(sock, cerrno)) return(-1);
    dctpoSocketLock(sock);
    if ((sock->state != DCTPSOCK_PARTOPEN) && (sock->state != DCTPSOCK_OPEN)) {
	if (cerrno) *cerrno = DCTP_EINVAL;
	dctpoSocketUnlock(sock);
	return(-1);
    }
    while ((*data = dctpiGetNextPacket(sock)) == NULL) {
	if ((sock->state != DCTPSOCK_PARTOPEN) && (sock->state != DCTPSOCK_OPEN)) {
	    if (cerrno) *cerrno = DCTP_ECONNREFUSED;
	    dctpoSocketUnlock(sock);
	    return(-1);
	}
	if (sock->flags & DCTPSOCK_O_NONBLOCK) {
	    if (cerrno) *cerrno = DCTP_EWOULDBLOCK;
	    dctpoSocketUnlock(sock);
	    return(-1);
	}
	dctpoSocketWait(sock);
    }
    dctpoSocketUnlock(sock);
    return((*data)->appdatalen);
}

/*
 * dctpaClose -- close a socket
 *
 * Inputs:  sock --   pointer to socket to close
 *          cerrno -- Pointer to errno-style variable for error details.
 *                    May be NULL if error details not needed.
 *
 * Returns: Zero (0) on success.
 *          -1 for error, with *cerrno set (if not NULL) to extended error.
 *          Extended errors mimick Unix call.
 */
int dctpaClose(dctpSocket *sock, int *cerrno) {
    int rval;

    if (badsock(sock, cerrno)) return(-1);
    dctpoSocketLock(sock);
    rval = dctpiCloseConnection(sock, cerrno);
    if (rval > 0) dctpoSocketUnlock(sock);
    return((rval < 0) ? -1 : 0);
}

int dctpiCloseConnection(dctpSocket *sock, int *cerrno) {
    dctpoLog(DCTPLOG_DEBUG, "dctpiCloseConnection: closing 0x%x, server %d\n",
	     (uint_t)sock, sock->server);
    switch (sock->state) {
    case DCTPSOCK_CLOSED:
    case DCTPSOCK_CREATED:
    case DCTPSOCK_LISTEN:
    case DCTPSOCK_BOUND:
	dctpiDestroySock(sock);
	return(0);
    case DCTPSOCK_REQUEST:
	dctpoStopTimer(&(sock->timer));
    case DCTPSOCK_RESPOND:
	sock->state = DCTPSOCK_CLOSED;
	dctpiSendReset(sock, NULL, DCTPRESET_ABORTED, 0, 0, 0, sock->encap);
	dctpiDestroySock(sock);
	return(0);
    case DCTPSOCK_PARTOPEN:
    case DCTPSOCK_OPEN:
	/* Set CLOSING timeout */
	sock->timeval = (sock->ccidSenderFuncs) ?
	    2*(sock->ccidSenderFuncs->getCurrentRTO(sock)) :
	    2*DCTP_INITRTO;
	/* Close CCIDs */
	if (sock->ccidSenderFuncs) sock->ccidSenderFuncs->destroyCcidInfo(sock);
	if (sock->ccidRcvrFuncs) sock->ccidRcvrFuncs->destroyCcidInfo(sock);
	if (sock->server) {
	    sock->state = DCTPSOCK_CLOSEREQ;
	    dctpiSendClosereq(sock);
	} else {
	    sock->state = DCTPSOCK_CLOSING;
	    dctpiSendClose(sock);

	}
	/* Socket will be destroyed once close completes */
	sock->destroy = 1;
	/* Start closing timer */
	dctpoStartTimer(&(sock->timer), sock->timeval, dctpiCloseReqTimeout, sock);
	return(1);
    case DCTPSOCK_CLOSEREQ:
    case DCTPSOCK_CLOSING:
    case DCTPSOCK_TIMEWAIT:
	/* These three in process of closing anyway, mark to destroy on close */
	sock->destroy = 1;
	return(1);
    default:
	dctpoLog(DCTPLOG_ERR, "dctpaClose: socket in unknown state: %d\n", sock->state);
	if (*cerrno) *cerrno = DCTP_EINVAL;
	dctpiDestroySock(sock);
	return(-1);
    }
}

/*
 * dctpaSetSockopt
 */
int dctpaSetSockopt(dctpSocket *sock, int opt, int mandatory,
		    void *opval, uint_t oplen, int *cerrno) {
    if (badsock(sock, cerrno)) return(-1);
    dctpoSocketLock(sock);
    switch (opt) {
    case DCTPSOCKOPT_LOCALCCID:
	dctpoMemset(sock->localFeatsDesired.ccid, 0, DCTPFEAT_CCIDSIZE);
	dctpoMemcpy(sock->localFeatsDesired.ccid, opval,
		    (oplen > DCTPFEAT_CCIDSIZE) ? DCTPFEAT_CCIDSIZE : oplen);
	sock->localFeatsState[DCTPFEAT_CCID] =
	    (mandatory) ? DCTPFEAT_MANUNSTABLE : DCTPFEAT_UNSTABLE;
	sock->sendFeatureNegotiation = 1;
	break;
    case DCTPSOCKOPT_REMOTECCID:
	dctpoMemset(sock->remoteFeatsDesired.ccid, 0, DCTPFEAT_CCIDSIZE);
	dctpoMemcpy(sock->remoteFeatsDesired.ccid, opval,
		    (oplen > DCTPFEAT_CCIDSIZE) ? DCTPFEAT_CCIDSIZE : oplen);
	sock->remoteFeatsState[DCTPFEAT_CCID] =
	    (mandatory) ? DCTPFEAT_MANUNSTABLE : DCTPFEAT_UNSTABLE;
	sock->sendFeatureNegotiation = 1;
	break;
    case DCTPSOCKOPT_LOCALSENDACKVECTOR:
	sock->localFeatsDesired.sendAckVector = *((uint8_t *)opval);
	sock->localFeatsState[DCTPFEAT_SENDACKVECTOR] =
	    (mandatory) ? DCTPFEAT_MANUNSTABLE : DCTPFEAT_UNSTABLE;
	sock->sendFeatureNegotiation = 1;
	break;
    case DCTPSOCKOPT_REMOTESENDACKVECTOR:
	sock->remoteFeatsDesired.sendAckVector = *((uint8_t *)opval);
	sock->remoteFeatsState[DCTPFEAT_SENDACKVECTOR] =
	    (mandatory) ? DCTPFEAT_MANUNSTABLE : DCTPFEAT_UNSTABLE;
	sock->sendFeatureNegotiation = 1;
	break;
    default:
	dctpoSocketUnlock(sock);
	if (cerrno) *cerrno = DCTP_ENOPROTOOPT;
	return(-1);
    }
    dctpoSocketUnlock(sock);
    return(0);
}
