/*
 * discard-client.c -- Discard client application for Lprocess port of dccp-tp
 *
 * Copyright (C) 2008 Tom Phelan
 *
 * This file is part of dccp-tp.
 *
 * Dccp-tp is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * Dccp-tp is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with dccp-tp.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Documentation and source code for dccp-tp is available at
 * http://www.phelan-4.com/dccp-tp/.
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include "dctpLprocessApi.h"

extern int h_errno;

#define DISCARD_DEFPORT   9

int main(int argc, char **argv) {
    int optc, sd, optval;
    int daemonize = 0;
    int type = DCTPLP_SOCKNAT;
    char *scodetxt = "SC:DCRD";
    int pcount, mlen = 1400, mxpcount = 100;
    dctpSockaddr addr;
    dctpSockaddrV4 *v4a = (dctpSockaddrV4 *)&addr;
    struct hostent *hent;
    char ccid = '3';
    char mbuf[1500];

    /* Set up defaults */
    addr.sa_family = AF_INET;
    addr.sa_port = htons(DISCARD_DEFPORT);
    v4a->sin_addr.s_addr = htonl(INADDR_LOOPBACK);
    for (sd = 0; sd < sizeof(mbuf); ++sd) {
	mbuf[sd] = sd;
    }

    /* Command line options may change things */
    while ((optc = getopt(argc, argv, "a:c:dp:")) > 0) {
	switch (optc) {
	case 'a':
	    if ((hent = gethostbyname(optarg)) == NULL) {
		printf("Can't covert %s to IPv4 address: %s\n", optarg, strerror(h_errno));
		return(-1);
	    }
	    v4a->sin_addr.s_addr = *((uint32_t *)(hent->h_addr_list[0]));
	    break;
	case 'c':
	    ccid = *optarg;
	    if ((ccid != '2') && (ccid != '3')) {
		printf("Invalid ccid %c\n", ccid);
		return(-1);
	    }
	    break;
	case 'd':
	    daemonize = 1;
	    break;
	case 'p':
	    sscanf(optarg, "%d", &mxpcount);
	    break;
	default:
	    printf("Invalid option %c\n", optc);
	    return(-1);
	}
    }
    if (daemonize) {
	if (daemon(0, 0) < 0) {
	    printf("Can't run as daemon: %s\n", strerror(errno));
	    return(-1);
	}
	openlog("dctp-dcard-client", LOG_PID, LOG_DAEMON);
	syslog(LOG_INFO, "discard-client running as daemon, pid %d\n", getpid());
    } else {
	openlog("dctp-dcard-client", LOG_PERROR | LOG_PID, LOG_USER);
	syslog(LOG_INFO, "discard-client running attached\n");
    }

    /* Open the socket */
    if ((sd = dctpMkSocket(addr.sa_family, type, scodetxt)) < 0) {
	syslog(LOG_ERR, "Can't open socket: %s\n", strerror(errno));
	return(-1);
    }
    if (ccid == '3') {
	/* Set up CCID 3 */
	optval = 3;
	if (dctpSetSockopt(sd, DCTPSOCKOPT_LOCALCCID, &optval, 1, 0) < 0) {
	    syslog(LOG_ERR, "Error setting local CCID: %s\n", strerror(errno));
	    dctpClose(sd);
	    return(-1);
	}
	if (dctpSetSockopt(sd, DCTPSOCKOPT_REMOTECCID, &optval, 1, 0) < 0) {
	    syslog(LOG_ERR, "Error setting remote CCID: %s\n", strerror(errno));
	    dctpClose(sd);
	    return(-1);
	}
	/* Turn off send ack vector */
	optval = 0;
	if (dctpSetSockopt(sd, DCTPSOCKOPT_LOCALSENDACKVECTOR, &optval, 1, 0) < 0) {
	    syslog(LOG_ERR, "Error setting local send ack vector: %s\n", strerror(errno));
	    dctpClose(sd);
	    return(-1);
	}
	if (dctpSetSockopt(sd, DCTPSOCKOPT_REMOTESENDACKVECTOR, &optval, 1, 0) < 0) {
	    syslog(LOG_ERR, "Error setting remote send ack vector: %s\n", strerror(errno));
	    dctpClose(sd);
	    return(-1);
	}
    }
    /* Connect to server */
    syslog(LOG_INFO, "Connecting to %s...\n",
	   inet_ntoa(*((struct in_addr *)&(v4a->sin_addr))));
    if (dctpConnect(sd, &addr, NULL, 0) < 0) {
	syslog(LOG_ERR, "Can't connect: %s\n", strerror(errno));
	dctpClose(sd);
	return(-1);
    }
    syslog(LOG_INFO, "Connected\n");
    /* Send some packets */
    for (pcount = 0; pcount < mxpcount; ++pcount) {
	*((int *)mbuf) = pcount;
	if (dctpSend(sd, mbuf, mlen) < 0) {
	    syslog(LOG_ERR, "Error sending packet %d: %s\n", pcount, strerror(errno));
	    break;
	}
	if (!daemonize) {
	    printf("%d\r", pcount);
	    fflush(stdout);
	}
    }
    if (!daemonize) printf("\n");
    syslog(LOG_INFO, "Terminating connection, sent %d packets\n", pcount);
    sleep(1);
    dctpClose(sd);
    return(0);
}
