/*
 * discard-server.c -- Discard server application for Lprocess port of dccp-tp
 *
 * Copyright (C) 2008 Tom Phelan
 *
 * This file is part of dccp-tp.
 *
 * Dccp-tp is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * Dccp-tp is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with dccp-tp.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Documentation and source code for dccp-tp is available at
 * http://www.phelan-4.com/dccp-tp/.
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <syslog.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "dctpLprocessApi.h"

#define DISCARD_DEFPORT   9

int main(int argc, char **argv) {
    int optc, sd, nsd, optval;
    int daemonize = 0;
    int type = DCTPLP_SOCKNAT;
    char *scodetxt = "SC:DCRD";
    int nconnects, pcount, bcount, mlen;
    dctpSockaddr addr, naddr;
    dctpSockaddrV4 *v4a = (dctpSockaddrV4 *)&addr;
    dctpSockaddrV4 *v4n = (dctpSockaddrV4 *)&naddr;
    char ccid = '3';
    char mbuf[1500];

    /* Set up defaults */
    addr.sa_family = AF_INET;
    addr.sa_port = htons(DISCARD_DEFPORT);
    v4a->sin_addr.s_addr = htonl(INADDR_ANY);

    /* Command line options may change things */
    while ((optc = getopt(argc, argv, "c:d")) > 0) {
	switch (optc) {
	case 'c':
	    ccid = *optarg;
	    if ((ccid != '2') && (ccid != '3')) {
		printf("Invalid ccid %c\n", ccid);
		return(-1);
	    }
	    break;
	case 'd':
	    daemonize = 1;
	    break;
	default:
	    printf("Invalid option %c\n", optc);
	    return(-1);
	}
    }
    if (daemonize) {
	if (daemon(0, 0) < 0) {
	    printf("Can't run as daemon: %s\n", strerror(errno));
	    return(-1);
	}
	openlog("dctp-dcard-server", LOG_PID, LOG_DAEMON);
	syslog(LOG_INFO, "discard-server running as daemon, pid %d\n", getpid());
    } else {
	openlog("dctp-dcard-server", LOG_PERROR | LOG_PID, LOG_USER);
	syslog(LOG_INFO, "discard-server running attached\n");
    }

    /* Open the socket */
    if ((sd = dctpMkSocket(addr.sa_family, type, scodetxt)) < 0) {
	syslog(LOG_ERR, "Can't open socket: %s\n", strerror(errno));
	return(-1);
    }
    if (ccid == '3') {
	/* Set up CCID 3 */
	optval = 3;
	if (dctpSetSockopt(sd, DCTPSOCKOPT_LOCALCCID, &optval, 1, 0) < 0) {
	    syslog(LOG_ERR, "Error setting local CCID: %s\n", strerror(errno));
	    dctpClose(sd);
	    return(-1);
	}
	if (dctpSetSockopt(sd, DCTPSOCKOPT_REMOTECCID, &optval, 1, 0) < 0) {
	    syslog(LOG_ERR, "Error setting remote CCID: %s\n", strerror(errno));
	    dctpClose(sd);
	    return(-1);
	}
	/* Turn off send ack vector */
	optval = 0;
	if (dctpSetSockopt(sd, DCTPSOCKOPT_LOCALSENDACKVECTOR, &optval, 1, 0) < 0) {
	    syslog(LOG_ERR, "Error setting local send ack vector: %s\n", strerror(errno));
	    dctpClose(sd);
	    return(-1);
	}
	if (dctpSetSockopt(sd, DCTPSOCKOPT_REMOTESENDACKVECTOR, &optval, 1, 0) < 0) {
	    syslog(LOG_ERR, "Error setting remote send ack vector: %s\n", strerror(errno));
	    dctpClose(sd);
	    return(-1);
	}
    }
    /* Bind to local address */
    if (dctpBind(sd, &addr) < 0) {
	syslog(LOG_ERR, "Can't bind address to socket: %s\n", strerror(errno));
	dctpClose(sd);
	return(-1);
    }
    /* Put in listen state */
    if (dctpListen(sd, 10) < 0) {
	syslog(LOG_ERR, "Error setting listen state: %s\n", strerror(errno));
	dctpClose(sd);
	return(-1);
    }
    /* Now wait for incoming connections */
    nconnects = 0;
    while ((nsd = dctpAccept(sd, &naddr)) >= 0) {
	/* Got new connection */
	++nconnects;
	syslog(LOG_INFO, "Accepted connection %d from %s\n", nconnects,
	       inet_ntoa(*(struct in_addr *)&(v4n->sin_addr)));
	pcount = 0;
	bcount = 0;
	/* Receive as long as they send */
	while ((mlen = dctpRecv(nsd, mbuf, sizeof(mbuf))) >= 0) {
	    ++pcount;
	    bcount += mlen;
#ifdef NEVER
	    if (!daemonize) {
		printf(".");
		fflush(stdout);
	    }
#endif
	}
	if (!daemonize) printf("\n");
	syslog(LOG_INFO, "Connection %d from %s, rcvd %d pkts, %d btyes: %s\n", nconnects,
	       inet_ntoa(*(struct in_addr *)&(v4n->sin_addr)), pcount, bcount,
	       strerror(errno));
	dctpClose(nsd);
    }
    syslog(LOG_INFO, "Accept failed: %s\n", strerror(errno));
    return(0);
}
