/*
 * echo-server.c -- Echo server application for Pthreads port of dccp-tp
 *
 * Copyright (C) 2008 Tom Phelan
 *
 * This file is part of dccp-tp.
 *
 * Dccp-tp is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * Dccp-tp is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with dccp-tp.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Documentation and source code for dccp-tp is available at
 * http://www.phelan-4.com/dccp-tp/.
 */

#include <pthread.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "dctpCore.h"
#include "dctpApi.h"
#include "dctpSupport.h"

#define ECHO_DEFPORT 7
#define ECHO_MAXTHREADS 16
#define ECHO_MAXPKTS    100000
#define ECHO_MAXPKTSIZE 1500

static uint16_t echoPort = ECHO_DEFPORT;
static uint_t maxthreads = ECHO_MAXTHREADS;
static uint_t threadcount = 0;
static pthread_cond_t threadcond = PTHREAD_COND_INITIALIZER;
static pthread_mutex_t threadmutex = PTHREAD_MUTEX_INITIALIZER;
static uint_t maxpkts = ECHO_MAXPKTS;

static void *echoServerThread(void *arg) {
    int sock = (int)arg;
    uint_t pkts = maxpkts, ecnt = 0;
    int plen;
    uint8_t buf[ECHO_MAXPKTSIZE];

    dctpoLog(DCTPLOG_DEBUG, "echoServerThread: starting echo-server thread [%ld]\n",
	     pthread_self());
    while (pkts--) {
	if ((plen = dccpRecv(sock, buf, sizeof(buf))) < 0) {
	    dctpoLog(DCTPLOG_ERR, "echoServerThread: can't read from socket: %d\n", errno);
	    break;
	}
	if (dccpSend(sock, buf, plen) < 0) {
	    dctpoLog(DCTPLOG_ERR, "echoServerThread: can't write to socket: %d\n", errno);
	    break;
	}
	++ecnt;
	/*dctpoLog(DCTPLOG_DEBUG, "echoServerThread: wrote packet %d\n", ecnt);*/
    }
    dctpoLog(DCTPLOG_DEBUG, "echoServerThread: closing, echoed %d packets\n", ecnt);
    dccpClose(sock);
    pthread_mutex_lock(&threadmutex);
    --threadcount;
    pthread_mutex_unlock(&threadmutex);
    return(NULL);
}

static void *echoServer(void *arg) {
    dctpSockaddrV4 addr;
    int sock, nsock, alen;
    pthread_attr_t pattr;
    pthread_t thread;

    pthread_attr_init(&pattr);
    pthread_attr_setdetachstate(&pattr, PTHREAD_CREATE_DETACHED);
    if ((sock = dccpSocket(DCTP_PF_INET, SOCK_DCCP_NAT, "SC:ECHO")) < 0) {
	dctpoLog(DCTPLOG_ERR, "echoServer: can't open socket: %d\n", errno);
	return(NULL);
    }
    addr.sin_family = DCTP_AF_INET;
    addr.sin_port = dctpoHtons(echoPort);
    addr.sin_addr.s_addr = dctpoHtonl(DCTP_INADDRANY);
    if (dccpBind(sock, (dctpSockaddr *)&addr, sizeof(addr)) < 0) {
	dctpoLog(DCTPLOG_ERR, "echoServer: can't bind socket: %d\n", errno);
	return(NULL);
    }
    if (dccpListen(sock, 32) < 0) {
	dctpoLog(DCTPLOG_ERR, "echoServer: can't listen: %d\n", errno);
	return(NULL);
    }
    alen = sizeof(addr);
    while ((nsock = dccpAccept(sock, (dctpSockaddr *)&addr, &alen)) >= 0) {
	dctpoLog(DCTPLOG_INFO, "echoServer: accepted connection from %s:%d\n",
		 inet_ntoa(*((struct in_addr *)&(addr.sin_addr))),
		 ntohs(addr.sin_port));
	pthread_mutex_lock(&threadmutex);
	while (threadcount >= maxthreads) {
	    dctpoLog(DCTPLOG_DEBUG, "echoServer: waiting to start thread...\n");
	    pthread_cond_wait(&threadcond, &threadmutex);
	}
	++threadcount;
	pthread_create(&thread, &pattr, echoServerThread, (void *)nsock);
	pthread_mutex_unlock(&threadmutex);
    }
    return(NULL);
}

void echoServerInit(void) {
    pthread_attr_t pattr;
    pthread_t thread;

    pthread_attr_init(&pattr);
    pthread_attr_setdetachstate(&pattr, PTHREAD_CREATE_DETACHED);
    pthread_create(&thread, &pattr, echoServer, NULL);
}
